
import os
import wrapt
from typing import Callable, TypeVar
from autoboot.applications import AutoBoot
from autoboot.annotation.component import component
from autoboot.process import get_yml_value

R = TypeVar("R")

def env(keypath: str) -> int | float | str | dict | list | None:
  """get env value from keypath, support find in `.env` and `.yml`."""
  
  if "." in keypath:
    value = get_yml_value(AutoBoot.get_config_data(), keypath=keypath)
    if not value:
      return None
    return value
  else:
    value = os.getenv(keypath)
    if value:
      return value
    value = AutoBoot.get_config_data().get(keypath)
    return value if value else None
    

def value(keypath: str) -> R:
  """get value from config file."""
  
  @wrapt.decorator
  def decorator(fn: Callable[..., R], instance, args, kwargs) -> R:
    return env(keypath) or fn(*args, **kwargs)
  return decorator
      

def value_component(keypath: str) -> R:
  """get value from config file and cache as component."""
  
  @wrapt.decorator
  @component(name=f"env[{keypath}]")
  def decorator(fn: Callable[..., R], instance, args, kwargs) -> R:
    return env(keypath) or fn(*args, **kwargs)
  return decorator



from autoboot.annotation.env import value_component


class ApplicationProperties:
  
  @value_component("autoboot.application.name")
  @staticmethod
  def app_name() -> str:
    return "app"

  @value_component("autoboot.application.module")
  @staticmethod
  def module() -> str:
    return "api"
  
  @value_component("autoboot.application.log.dir")
  @staticmethod
  def log_dir() -> str:
    return "logs"
  
  @value_component("autoboot.application.log.max_size")
  @staticmethod
  def log_max_size() -> str:
    return "100 MB"

  @value_component("autoboot.application.log.retention")
  @staticmethod
  def log_retention() -> str:
    return "30 days"



from concurrent.futures import ThreadPoolExecutor
from autoboot.meta import Component
from autoboot import ServerProperties
from typing import Callable, Any


class ThreadPoolTask(Component):
  
  def __init__(self):
    self.executor = ThreadPoolExecutor(max_workers=ServerProperties.thread_max_workers())
    
  def worker_size(self):
    return self.executor._work_queue.qsize()
    
  def submit(self, fn: Callable[..., Any]):
    executing_tasks = self.worker_size()
    max_workers = ServerProperties.thread_max_workers()
    rejected = ServerProperties.thread_rejected()
    
    if executing_tasks > max_workers:
      if rejected == "abort":
        raise RuntimeError("ThreadPoolExecute: Too many threads executing")
      elif rejected == "discard":
        return
      
    self.executor._work_queue.put(fn)
        



from typing import Self
from autoboot.annotation.component import comp_from_cache


class Component:
  """Cache component and register it in IoC."""

  def __new__(cls) -> Self:
    cls.__component_name__ = cls.__name__
    _parent = super()
    return comp_from_cache(cls.__name__, lambda: _parent.__new__(cls))
  



import abc


class AppPlugin(abc.ABC):
  
  @abc.abstractmethod
  def install(self) -> None:
    pass
  
  @abc.abstractmethod
  def env_prepared(self) -> None:
    pass
  
  @abc.abstractmethod
  def app_started(self) -> None:
    pass



from autoboot import ApplicationProperties, AutoBoot, AutoBootConfig


def testProperties():
  autoboot = AutoBoot(config=AutoBootConfig(config_dir="./tests/test_ioc_properties"))
  autoboot.run()
  
  assert(ApplicationProperties.app_name() == "test-app-dev")



import loguru
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, TypeVar
from autoboot.args import get_env_name
from autoboot.logging import Logging
from autoboot.plugin import AppPlugin
from autoboot.process import load_env_file, load_yaml_file

# define AutoBoot class type
AppType = TypeVar("AppType", bound="AutoBoot")

@dataclass
class AutoBootConfig:
  """AutoBoot global config"""
  
  # show detail log
  verbose: bool = False
  # config dir path which relative to the current project
  config_dir: str = "."
  # yaml config name, it support `.yml` and `.yaml`
  yml_config: str = "autoboot"
  
  

class AutoBoot(object):
  """create instance of AutoBoot."""
  
  _instance = None
  _init_flag = False
  
  def __new__(cls, config: Optional[AutoBootConfig]) -> AppType:
    # create instance only once
    if cls._instance is None:
        cls._instance = super().__new__(cls)
    return cls._instance
  
  def __init__(self: AppType, config: Optional[AutoBootConfig] = AutoBootConfig()) -> None:
    if AutoBoot._init_flag is True:
      return
    self.config = config
    AutoBoot.logger: loguru.Logger = None
    self._app_plugins: List[AppPlugin] = []
    self._components: List[Tuple[str, Any]] = []
    AutoBoot._config_data = dict[str, Any]
    AutoBoot._init_flag = True
    
  @classmethod
  def instance(cls: AppType) -> AppType:
    return cls._instance
  
  @staticmethod
  def get_config_data() -> dict[str, Any]:
    return AutoBoot._config_data
    
  
  @property
  def app_plugins(self) -> List[AppPlugin]:
    return self._app_plugins
  
  @app_plugins.setter
  def app_plugins(self, app_plugins: List[AppPlugin]) -> None:
    self._app_plugins = app_plugins 
  
  @property
  def components(self) -> List[Tuple[str, Any]]:
    return self._components
    
  def apply(self: AppType, app_plugin: AppPlugin) -> None:
    app_plugin.install(self)
    self._app_plugins.append(app_plugin)
  
  def run(self: AppType):
    # config environment variables
    env_name = get_env_name()
    
    # load .env
    load_env_file(self.config.config_dir, env_name)
    
    # load yml
    AutoBoot._config_data = load_yaml_file(self.config.config_dir, self.config.yml_config)
    
    # config logger
    from autoboot.application_properties import ApplicationProperties
    AutoBoot.logger = Logging(
      app_name=ApplicationProperties.app_name(),
      module=ApplicationProperties.module(),
      log_dir=ApplicationProperties.log_dir(),
      max_size=ApplicationProperties.log_max_size(),
      retention=ApplicationProperties.log_retention(),
      env_name=env_name
    ).logger
    
    AutoBoot.logger.info(f"application finish load env: {env_name}")
    
    for app_plugin in self._app_plugins:
      app_plugin.env_prepared()

    for app_plugin in self._app_plugins:
      app_plugin.app_started()
      
    AutoBoot.logger.info("application started!")



from autoboot.annotation.env import value_component

class ServerProperties:

  @value_component("server.threads.max")
  @staticmethod
  def thread_max_workers() -> int:
    """The max workers is 31"""
    return 28
  
  @value_component("server.threads.rejected")
  @staticmethod
  def thread_rejected() -> str:
    """Available reject strategy: abort, discard, wait"""
    return "wait"


from setuptools import find_packages, setup

VERSION = '0.5.0'

with open('README.md', 'r', encoding='utf-8') as fp:
    long_description = fp.read()

setup(
    name='autoboot',
    version=VERSION,
    author='yizzuide',
    author_email='fu837014586@163.com',
    description='IoC with auto config framework',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/yizzuide/autoboot',
    keywords=['IoC', "event", 'auto config'],
    packages=find_packages(exclude=['tests*']),
    include_package_data=True,
    install_requires=[
        'python-dotenv>=1.0.0',
        'filetype>=1.2.0',
        'loguru>=0.7.0',
        'pyyaml>=6.0.1',
        'wrapt>=1.14.1',
        'pydantic>=1.10.12',
        'result>=0.14.0'
    ],
    extra_require={
        'web': ['fastapi>=0.70.1', 'uvicorn>=0.16.0']
    },
    tests_require=[
        'pytest>=6.2.0',
        'pytest-cov>=2.10.0'
    ],
    python_requires='>=3.8',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License'
    ],
)